from copy import deepcopy
from dataclasses import dataclass

from watermill.expressions import get_field
from watermill.stream_join import get_tree_pairs
from watermill.stream_join import join_streams
from watermill.stream_join import JoinTree
from watermill.stream_join import JoinTreeNode
from watermill.stream_join import JoinTreePair
from watermill.stream_join import JoinWith


@dataclass
class Foo:
    pass


@dataclass
class Bar:
    pass


@dataclass
class Baz:
    pass


def test_build_simple_join():
    left_expression = get_field('field')
    right_expression = get_field('field')
    join = join_streams(
        Foo,
        JoinWith(
            with_type=Bar,
            left_expression=left_expression,
            right_expression=right_expression,
        ),
    )
    assert isinstance(join, JoinTree)
    assert join.root_node == JoinTreeNode(user_type=Foo)
    assert get_tree_pairs(join, join.root_node) == [JoinTreePair(
        left_node=JoinTreeNode(user_type=Foo),
        right_node=JoinTreeNode(user_type=Bar),
        left_expression=left_expression,
        right_expression=right_expression,
    )]


def test_build_multi_join():
    left_expression = get_field('field')
    right_expression = get_field('field')

    second_left_expression = get_field('field')
    second_right_expression = get_field('field')

    join = join_streams(
        Foo,
        JoinWith(
            with_type=Bar,
            left_expression=left_expression,
            right_expression=right_expression,
        ),
        JoinWith(
            with_type=Baz,
            left_expression=second_left_expression,
            right_expression=second_right_expression,
        )
    )

    assert isinstance(join, JoinTree)
    assert join.root_node == JoinTreeNode(user_type=Foo)
    assert get_tree_pairs(join, join.root_node) == [JoinTreePair(
        left_node=JoinTreeNode(user_type=Foo),
        right_node=JoinTreeNode(user_type=Bar),
        left_expression=left_expression,
        right_expression=right_expression,
    ), JoinTreePair(
        left_node=JoinTreeNode(user_type=Foo),
        right_node=JoinTreeNode(user_type=Baz),
        left_expression=second_left_expression,
        right_expression=second_right_expression,
    )]


def test_build_chained_join():
    left_expression = get_field('field')
    right_expression = get_field('field')

    second_left_expression = get_field('field')
    second_right_expression = get_field('field')

    nested_join = join_streams(
        Bar,
        JoinWith(
            with_type=Baz,
            left_expression=second_left_expression,
            right_expression=second_right_expression,
        )
    )

    join = join_streams(
        Foo,
        JoinWith(
            with_type=nested_join,
            left_expression=left_expression,
            right_expression=right_expression,
        ),
    )

    assert isinstance(join, JoinTree)
    assert join.root_node == JoinTreeNode(user_type=Foo)
    assert get_tree_pairs(join, join.root_node) == [JoinTreePair(
        left_node=JoinTreeNode(user_type=Foo),
        right_node=JoinTreeNode(user_type=Bar),
        left_expression=left_expression,
        right_expression=right_expression,
    )]
    assert get_tree_pairs(join, JoinTreeNode(user_type=Bar)) == [JoinTreePair(
        left_node=JoinTreeNode(user_type=Bar),
        right_node=JoinTreeNode(user_type=Baz),
        left_expression=second_left_expression,
        right_expression=second_right_expression,
    )]

    def test_build_multi_chained_join():
        #
        #        Baz
        #       /
        #     Bar
        #    /   \ Quux
        # Foo
        #    \
        #     Qwe
        #      \
        #       Asd

        Quux, Qwe, Asd = (deepcopy(cls) for cls in [Bar] * 3)
        e = get_field('field')

        bbq_join = join_streams(
            Bar,
            JoinWith(
                with_type=Baz,
                left_expression=e,
                right_expression=e
            ),
            JoinWith(
                with_type=Quux,
                left_expression=e,
                right_expression=e
            )
        )

        qa_join = join_streams(
            Qwe,
            JoinWith(
                with_type=Asd,
                left_expression=e,
                right_expression=e
            )
        )

        join = join_streams(
            Foo,
            JoinWith(
                with_type=bbq_join,
                left_expression=e,
                right_expression=e
            ),
            JoinWith(
                with_type=qa_join,
                left_expression=e,
                right_expression=e
            )
        )

        assert join.root_node == JoinTreeNode(user_type=Foo)
        assert get_tree_pairs(join, join.root_node) == [JoinTreePair(
            left_node=JoinTreeNode(user_type=Foo),
            right_node=JoinTreeNode(user_type=Bar),
            left_expression=e,
            right_expression=e,
        ), JoinTreePair(
            left_node=JoinTreeNode(user_type=Foo),
            right_node=JoinTreeNode(user_type=Qwe),
            left_expression=e,
            right_expression=e,
        )]

        assert get_tree_pairs(join, JoinTreeNode(user_type=Bar)) == [JoinTreePair(
            left_node=JoinTreeNode(user_type=Bar),
            right_node=JoinTreeNode(user_type=Baz),
            left_expression=e,
            right_expression=e,
        ), JoinTreePair(
            left_node=JoinTreeNode(user_type=Bar),
            right_node=JoinTreeNode(user_type=Quux),
            left_expression=e,
            right_expression=e,
        )]

        assert get_tree_pairs(join, JoinTreeNode(user_type=Qwe)) == [JoinTreePair(
            left_node=JoinTreeNode(user_type=Qwe),
            right_node=JoinTreeNode(user_type=Asd),
            left_expression=e,
            right_expression=e,
        )]



import enum
from dataclasses import dataclass
from typing import Any
from typing import Callable
from typing import Type
from typing import TypeVar


StreamType = TypeVar('StreamType')
JoinedStreamType = TypeVar('JoinedStreamType')


class EndOfStreamError(Exception):
    def __init__(
            self,
            stream_type: Type[StreamType],
    ):
        super().__init__()
        self._stream_type = stream_type


class StreamTimeoutError(Exception):
    def __init__(
            self,
            stream_type: Type[StreamType],
    ):
        super().__init__()
        self._stream_type = stream_type


class RightJoinedStreamRanAhead(Exception):
    def __init__(self, stream_type: Type[JoinedStreamType]):
        super().__init__()
        self.stream_type = stream_type


@dataclass
class EndOfStream:
    eos__: bool = True


class StreamsShift(enum.Enum):
    Less = 'Less'
    Equal = 'Equal'
    Greater = 'Greater'


class MessageBroker:
    def send(self, item_type: Type[StreamType], item: dict):
        raise NotImplementedError()

    def send_eos(self, item_type: Type[StreamType]):
        raise NotImplementedError()

    def get_elements(
            self,
            element_type: Type[StreamType],
            comparator: Callable[[Any, Any], StreamsShift] = None
    ) -> dict:
        raise NotImplementedError()

    def commit(self):
        pass



import os
from dataclasses import dataclass
from pathlib import Path
from typing import NoReturn

from botox import Injector

from watermill.message_brokers.json_file_message_broker import JsonFileMessageBroker
from watermill.mill import WaterMill


class QuantityCalculator:
    def __init__(self):
        self.quantity = 0

    def add(self, quantity: int):
        self.quantity += quantity


@dataclass(frozen=True, eq=True)
class GoodsItem:
    title: str
    quantity: int
    price: int


def sum_good_quantities(goods_item: GoodsItem, side_effect: QuantityCalculator) -> NoReturn:
    side_effect.add(goods_item.quantity)


def single_stream_example():
    injector = Injector()
    injector.prepare(QuantityCalculator, QuantityCalculator())

    message_broker = JsonFileMessageBroker({
        GoodsItem: Path(os.path.join(os.path.dirname(__file__), 'samples/goods.json'))
    })

    mill = WaterMill(
        broker=message_broker,
        process_func=sum_good_quantities,
        stream_cls=GoodsItem,
        injector=injector
    )
    mill.run()

    calc = injector.deliver(QuantityCalculator)
    assert calc.quantity == 8


if __name__ == "__main__":
    single_stream_example()



import json
import math
from dataclasses import asdict
from dataclasses import dataclass
from pathlib import Path
from time import sleep
from typing import List
from typing import Mapping

from kafka import KafkaAdminClient
from kafka import KafkaConsumer
from kafka import KafkaProducer
from kafka.admin import NewTopic

from watermill.expressions import get_field
from watermill.message_brokers.kafka_message_broker import KafkaMessageBroker
from watermill.message_brokers.message_broker import EndOfStream
from watermill.mill import WaterMill
from watermill.stream_join import JoinWith
from watermill.stream_join import join_streams


@dataclass(frozen=True, eq=True)
class RobotLocation:
    time: int
    x: int
    y: int


@dataclass(frozen=True, eq=True)
class RobotLocationNextStep:
    time: int
    x: int
    y: int


@dataclass(frozen=True, eq=True)
class RobotSpeed:
    time_from: int
    time_to: int
    moment: float
    speed: float


def calculate_robot_speed(prev_location: RobotLocation, next_locations: Mapping[RobotLocation, List[RobotLocationNextStep]]) -> RobotSpeed:
    next_location = next_locations[prev_location][0]
    for x in range(10):
        yield RobotSpeed(
            time_from=prev_location.time,
            time_to=next_location.time,
            moment=int((0.1 * x) * 10) / 10 * (next_location.time - prev_location.time) + prev_location.time,
            speed=math.sqrt(math.pow(next_location.x - prev_location.x, 2) + math.pow(next_location.y - prev_location.y, 2))
        )


def single_input_multiple_results():
    message_broker = KafkaMessageBroker(
        topic_names={
            RobotLocation: 'robot-locations',
            RobotLocationNextStep: 'robot-locations',
            RobotSpeed: 'robot-speed',
        },
        kafka_url='kafka:9092'
    )

    mill = WaterMill(
        broker=message_broker,
        process_func=calculate_robot_speed,
        join_tree=join_streams(
            RobotLocation,
            JoinWith(
                with_type=RobotLocationNextStep,
                left_expression=get_field('time') + 10,
                right_expression=get_field('time')
            ),
        ),
    )

    produce_goods_stream()

    mill.run()

    print('Fetching data from resulting stream')
    result_consumer = KafkaConsumer(
        bootstrap_servers='kafka:9092',
        group_id='result-consumer',
        auto_offset_reset='earliest',
        enable_auto_commit=True,
    )
    result_consumer.subscribe(['robot-speed'])
    for result in result_consumer:
        result_item = json.loads(result.value.decode('utf8'))
        print(json.dumps(result_item))
        if 'eos__' in result_item:
            break


def produce_goods_stream():
    recreate_topics(['robot-locations', 'robot-speed'])
    print('Sending data to input stream')

    kafka_producer = KafkaProducer(bootstrap_servers='kafka:9092')

    goods_file_path = Path(__file__).resolve().parent / 'samples' / 'robot-locations.json'
    with goods_file_path.open('r') as f:
        goods = json.load(f)

    for goods_item in goods:
        print(goods_item)
        kafka_producer.send('robot-locations', value=json.dumps(goods_item).encode('utf8'))

    kafka_producer.send('robot-locations', value=json.dumps(asdict(EndOfStream())).encode('utf8'))
    kafka_producer.flush()


def recreate_topics(topics: List[str]):
    kafka_admin_client = KafkaAdminClient(
        bootstrap_servers='kafka:9092',
    )
    try:
        kafka_admin_client.delete_topics(topics)
    except Exception as exc:
        print(exc)
    sleep(0.1)
    try:
        kafka_admin_client.create_topics([NewTopic(name=topic, num_partitions=1, replication_factor=1) for topic in topics])
    except Exception as exc:
        print(exc)


if __name__ == "__main__":
    single_input_multiple_results()



import math

from watermill.expressions import Expression
from watermill.expressions import get_field


def test_get_field_expression():
    expr = get_field('field')
    assert isinstance(expr, Expression)

    result = expr({'field': 123})
    assert result == 123


def test_get_item_expression():
    expr = get_field('field')['sub']

    result = expr({
        'field': {
            'sub': 12.
        }
    })
    assert result == 12.


def test_add_expression():
    expr = get_field('field') + 15
    result = expr({'field': 2})
    assert result == 17



def test_div_expression():
    expr = get_field('field') // 3
    result = expr({'field': 15})
    assert result == 5


def test_sub_expression():
    expr = get_field('field') - 12.4
    result = expr({'field': 22.4})
    assert math.isclose(result, 10)


def test_mul_expression():
    expr = get_field('field') * 7
    result = expr({'field': 11})
    assert result == 77


def test_mod_expression():
    expr = get_field('field') % 3
    result = expr({'field': 16})
    assert result == 1


def test_multi_field_expressions():
    expr = get_field('left') % get_field('right') + 1
    result = expr({
        'left': 16,
        'right': 3
    })
    assert result == 2
